/* 
* File:			Vmpass.h
* DATE:     05-15-2007
* AUTHOR:   Lan Bai
* DESCRIPTION: This file declares stuff that is used in Vmpass.cpp.
*
* CHANGELOG: 
*/

#ifndef VMPASS_H
#define VMPASS_H

#include "llvm/Pass.h"
#include "llvm/Function.h"
#include "llvm/Instruction.h"
#include "llvm/Instructions.h"
#include "llvm/BasicBlock.h"
#include "llvm/Module.h"
#include "llvm/Transforms/Utils/BasicBlockUtils.h"
#include "llvm/Constants.h"
#include "llvm/ADT/Statistic.h"
#include "llvm/GlobalVariable.h"
#include "llvm/DerivedTypes.h"
#include "llvm/Target/TargetData.h"
#include <iostream>

	using namespace llvm;	
	using namespace std;

	Statistic <> BigGlobalNum ("vmpass",
		"Number of globals changed to virtual memory object");
	Statistic <> VmallocNum ("vmpass", "Number of calls to vm_kma_malloc");
	Statistic <> WriteNum ("vmpass", "Number of calls to write_handle");
	Statistic <> ReadNum ("vmpass", "Number of calls to read_handle");
	Statistic <> GlobalSize ("vmpass" , "Total size of globals");


	class Vmpass : public ModulePass {

	public:
		Vmpass();
 		virtual bool runOnModule(Module & M);
		virtual void getAnalysisUsage(AnalysisUsage & AU) const{
			AU.addRequired<TargetData>();
		}

	private:
		struct NewGInfo{
			GlobalValue *NewGlobal;
			int GSize;
		};

		/* MEMMU is used when total memory requirement is > total_size (byte). */
		const int total_page;

		/* Globals with size larger >= single_size (byte) is transformed by MEMMU. */
		const int single_size;

		/* Pagesize */
		const int page_size;

		/* Estimated compression ratio. */
		const float comp_ratio;

		/* Page number of compressed area. */
		int cm_page;

		/* Page number of uncompressed area. */
		int pm_page;

		/* Size of compressed area. */
		int cm_size;

		/* Size of uncompressed area. */
		int pm_size;

		int required_size;

		int required_page;

		/* first is the big global in original code
		second is new global <uint> created, and its size */
		map<GlobalValue *, NewGInfo> BigGlobals;

		/* first is load or store to old big globals
		second is call to read_handle/write_handle */
		map<Instruction *, Instruction *> InstMap;

		//first is GetelementPtrInst referencing big globals
  	//second is new instructin to get the virtual address
		map<Instruction *, Instruction *> GepMap;

		//get the total size of globals
		int SizeofGlobals(Module &M, const TargetData &TD);

		//divide memory to compressed and uncompressed regions
		void DivideMem();

		//create new globals based on previous ones, expand name by 'vm'
		void ChangeGlobals(GlobalValue * GV, int bigglobalsize);

		bool isBigGlobal (GlobalValue * GV);

		void getGepOffset(GetElementPtrInst * GEP, 
			Type::TypeID * Basetypeid, const TargetData &TD);

		uint64_t getOffset (ConstantExpr * CE, Type::TypeID * Basetypeid) const;

		GlobalValue *findNewg(GlobalValue * GV);

 		void Replaceload(Type::TypeID * basetypeid, Instruction * II,
			GlobalValue * GVa, uint64_t voffset);

		void Replacestore(Type::TypeID * basetypeid, Instruction * II,
			GlobalValue * GVa, uint64_t voffset);

		//replace all big globals with new ones
		//insert call to vm_malloc for each
		void Replaceglobals(Module * M);

		//update data flow of load from big globals
		void Replaceinsts(Module * M);

		void Replacegmpload(Instruction * LI);

		void Replacegmpstore(Instruction * SI);

		void Replacegmpgep(GetElementPtrInst *GEP, Instruction *BI, const
			TargetData &TD);

		void ReplaceGepUses(Instruction * II, const TargetData &TD);
	};

#endif

