  /*
 *  Filename:     kma_rm.h
 *  Purpose:      header file for kma_rm.c
 *  Description:  Resource Map Allocator, we adopt "Best fit" policy here,
 *                always allocate memory from the smallest region that is large
 *                enough to satisfy the request. This has the drawback that it
 *                might leave several free regions that are too small to be useful.                
 *                
 *  Author:       Lei Yang
 *  Date Created: 2004-11-07
 *
 *  CHANGELOG:
 *       Author:  Lei Yang
 *       Date:    2004-11-07
 *       Changes: Created
 *
 *       Author:  Lan Bai
 *       Date:    2005-12-27
 *       Changes: There is only one chunck, which is my virtual memory area
 *                so no need of pagelist. Change pagelist to vminfo
 *                Change all the pointers to offset
 *
 *       Author:  Lan Bai
 *       Date:    2006-02-04
 *       Changes: Introduce PM_INFO_RESERVE,CM_INFO_RESERVE,ad_size_t
 *
 */
 
#ifndef KMA_RM_H_
#define KMA_RM_H_

#include <stdint.h>
#include <stdlib.h>
#include <stdio.h>
#include "solu1.h"
#define DEBUG 1

#define MAX_CM_OBJ 54
#define MAX_VM_OBJ 6 

extern uint8_t cm_mem_area[];
extern uint8_t pm_mem_area[];
extern uint8_t *pm_start;
extern uint8_t *cm_start;
extern uint8_t *cbuf;
extern struct vminfo *pm_area;
extern struct vminfo *cm_area;
extern struct objinfo **pm_res_fp; /* pointer to the valid address left in reserved area */
extern struct objinfo **cm_res_fp; /* pointer to the valid address left in reserved area */
struct cm_objinfo {
  uint8_t * caddr;
  uint16_t csize;
  struct cm_objinfo *cnext;        /* next free chunk */
};

extern struct cm_objinfo cm_meminfo[MAX_CM_OBJ];	

typedef uint16_t ad_size_t;    /* when total vm and cm size is less than 32k */

struct vm_objinfo {
  uint16_t vaddr;
  uint16_t vsize;
  struct vm_objinfo *vnext;        /* next free chunk */
};

extern ad_size_t vm_kma_malloc(ad_size_t len);
extern uint8_t cm_kma_malloc(ad_size_t len);
extern void c_kma_free(uint8_t freeid);
extern void vm_initial();
extern void cm_initial();

#endif /* __KMA_RM_H__ */


