/*
 *  Filename:     solu1.c
 *  Purpose:      Inplement solu1.c
 *  Description:  SW-based Memory Management for applications which require more
 *                memory than RAM. We adopt LRU policy for page replacement here.
 *                Maintain a list of referenced pages, always swap out from the
 *                head of the list. 
 *                                
 *                
 *  Author:       Lan Bai
 *  Date Created: 2005-10-9
 *
 *  CHANGELOG:
 *  
 *       Author:  Lan Bai
 *       Date:    2006-02-04
 *       Changes: some bugs were discovered! in swap_out,set_free...
 *                
 *       Author:  Lan Bai
 *       Date:    2006-02-07
 *       Changes: the way to compute offset should be changed when pagesize changes.  
 *                so add a OFFSET_MASK
 *                but when use rle, one decompressed page is of size 258!
 *           
 *       Author:  Lan Bai
 *       Date:    2006-03-29
 *       Changes: updated with kma_rm.c, reduced some redundant computations,
 *                 added memory coalesce.
 *      
 *       Author:  Lan Bai
 *       Date:    2007-05-20
 *       Changes: globals related to compressed and uncompressed sizes are
 *								defined by vmpass.
 *                
 *                
 */

#include <stdlib.h>
#include <string.h>
#include "solu1.h"
#include "dcomp.h"

//uint8_t map[MAX_VM_PAGE];
//uint8_t freepages[MAX_PM_PAGE];
//struct lru_list free_list[MAX_PM_PAGE+1];
struct lru_list *free_head;
struct lru_list list_head;
struct lru_list *list_tail;
size_t dest_len;

#ifdef DEBUG
uint8_t swap_out_num = 0;
uint8_t swap_in_num = 0;
uint16_t access_num = 0;
float comp_ratio = 0;
#endif
/*************************** Function Declaration ****************************/
static void swap_in(handle);
static uint8_t swap_out();
static void decompress(handle,uint8_t,int);
static void compress(handle);
static void add_to_list(handle);
static void replace_to_list(handle);
static void setfree(struct lru_list *);
#ifdef DEBUG
void print_list();
#endif
/************************* Function Implementation****************************/
#if 0
int main() {
  uint16_t data_array,i;
  uint16_t x;
  initial();
  data_array = vm_kma_malloc(2500);
	for(i = 0; i < 1000; ++i) {
		write_handle16(data_array + i*2, 100);
  }
  for(i = 0; i < 1000; ++i) {
    x = read_handle16(data_array + i*2);
    if(x != 100) {
			printf("!!!!!!!!!! error: data is not as expected\n");
    }
  }
	printf("sucess!\n");
}
#endif
/*****************************************************************************/
static void compress(handle h) {
		uint8_t * src = pm_start + (uint16_t)(map[h]<<AD_SHIFT);  
		uint8_t com_fp_num;  
		uint8_t *dest;

//    pbpm_compress(src,PAGE_SIZE,cbuf,&dest_len, 0);
    delta_compress(src,PAGE_SIZE,cbuf,&dest_len, 0);

	if(dest_len == 0) {
#ifdef DEBUG
		printf("size after compression: %d is bigger!\n",dest_len);
#endif
   com_fp_num = cm_kma_malloc(PAGE_SIZE);
   dest = cm_meminfo[com_fp_num].caddr;
   memcpy(dest,src,PAGE_SIZE);
   map[h] = (MAX_PM_PAGE + com_fp_num)|0x80; 
#ifdef DEBUG
    printf("handle %d is uncompressable, copy to comp_area: %d\n",h,dest - cm_start);
#endif  
	} else {
    	com_fp_num = cm_kma_malloc(dest_len);
    	dest = cm_meminfo[com_fp_num].caddr;
    	memcpy(dest,cbuf,dest_len);
    	map[h] = MAX_PM_PAGE + com_fp_num;

#ifdef DEBUG
    	printf("finish compress handle %d to comp_area: %d\n",h, dest - cm_start);
    	printf("dest_len is : %d , compression ratio is %f\n",dest_len,(float)dest_len/(float)PAGE_SIZE);

    	if(comp_ratio != 0) {
      	comp_ratio = (comp_ratio + dest_len/(float)PAGE_SIZE)/2;
    	} else {
      	comp_ratio = dest_len/(float)PAGE_SIZE;
    	}
#endif
   	}
}

/*****************************************************************************/
static void decompress(handle h,uint8_t oldp,int decomp) {
#ifdef DEBUG
    printf("start decompress handle %d\n",h);
#endif

    struct cm_objinfo *resfp = &cm_meminfo[oldp];
    uint8_t *src = resfp->caddr;
    uint8_t *dest = pm_start + (uint16_t)(map[h]<<AD_SHIFT);

    if(!decomp) {
       memcpy(dest,src,PAGE_SIZE);
    } else {
//    	pbpm_decompress(src,resfp->csize,dest,&dest_len,0);

      delta_decompress(src,resfp->csize,dest,&dest_len,NULL);
    	if(dest_len != PAGE_SIZE) {
#ifdef DEBUG
      	printf("Error: dest_len changed after decompression! It's %d now\n",dest_len);
#endif
    	}
    }
#ifdef DEBUG
    printf("finish decompress from com_area %d to %d\n",resfp->caddr - cm_start,map[h]);
#endif
    c_kma_free(oldp);

}

/*****************************************************************************/      
static uint8_t swap_out() {
  handle s_h = list_head.next_p->list_handle; 
  uint8_t valid_page; 
  struct lru_list * freep = 0;
#ifdef DEBUG
  printf("swap out handle %d\n",s_h);
#endif
	valid_page = map[s_h];
  compress(s_h);
  
	freep = list_head.next_p;
  setfree(freep);
	list_head.list_handle--;
#ifdef DEBUG
  swap_out_num++;
#endif
//  call Leds.redToggle();
	return valid_page;
}   

/*****************************************************************************/
static void swap_in(handle h) {
  uint8_t i;
  int decomp;
  uint8_t freepage = NEVERUSE;
  uint8_t old_place;
#ifdef DEBUG
   printf("swap in handle %d\n", h);
#endif

  if((map[h]&0x80)==0) {
     decomp = 1;
     old_place = map[h] - MAX_PM_PAGE;
  } else {
    decomp = 0;
  	old_place = (map[h]&0x7f) - MAX_PM_PAGE;
  }
	//first,see whether there is a free page
  for(i = 0; i < MAX_PM_PAGE; ++i) {
    if(freepages[i] != NEVERUSE) {
      freepage = freepages[i];
      freepages[i] = NEVERUSE;
      break;
    }
  }
  if(freepage != NEVERUSE) {
    map[h]= freepage;
  } else {
    map[h] = swap_out();
  }
  decompress(h,old_place,decomp);

#ifdef DEBUG
  swap_in_num++;
  printf("now the lru list is: \n");
  print_list();
#endif
}

/*****************************************************************************/
uint16_t vir_to_phy(uint16_t  vir_addr) {
    uint8_t vir_page = vir_addr >> AD_SHIFT;
    uint16_t vir_offset = vir_addr & (uint16_t)OFFSET_MASK;
    uint8_t phy_page = map[vir_page];
    return (uint16_t)(phy_page<<AD_SHIFT)+vir_offset;
}

/*****************************************************************************/
// insert freeknot to the head of the freelist, free_list's count++
static void setfree(struct lru_list * freeknot) {
    freeknot->next_p->pre_p = freeknot->pre_p;
    freeknot->pre_p->next_p = freeknot->next_p;
    freeknot->next_p = free_head->next_p;
    free_head->next_p = freeknot;
    free_head->pre_p = NULL;
}

/*****************************************************************************/
static struct lru_list * getfree() {
    struct lru_list * p = free_head;
		if(free_head->next_p != NULL) {
			free_head = free_head->next_p;
    } else {
    }
    return p;
}

/*****************************************************************************/
static void add_to_list(handle h) {
    struct lru_list * p = getfree();
    p->list_handle = h;
    list_tail->next_p = p;
    p->pre_p = list_tail;
    list_tail = p;
    p->next_p = list_tail;
    list_head.list_handle++;
#ifdef DEBUG
    printf("add  handle %d to list\n",h);
#endif
}

/*****************************************************************************/
// if page h is already in lru-list, replace it to the end
// otherwise, append it to the end of the list
static void replace_to_list(handle h) {
    struct lru_list * p = list_tail;
    for(;p != &list_head; p = p->pre_p) {
	    if (p->list_handle == h) {
	    break;
	    }
    }
    if(p != &list_head) {
      if(p!= list_tail) {
	     p->pre_p->next_p = p->next_p;
	     p->next_p->pre_p = p->pre_p;
    	 list_tail->next_p = p;
       	p->pre_p = list_tail;
	     list_tail = p;
        p->next_p = list_tail;
	   } 
    } else { 
	    add_to_list(h);
    }
}

/*****************************************************************************/
#ifdef DEBUG
void print_list() {
    struct lru_list * p = list_head.next_p;
    printf("list: ");
    if(!p) {
    	for(;p!=list_tail;p=p->next_p) {
	    	printf("%d -> ",p->list_handle);
    	}
		} else {
			printf("empty\n");
		}
    printf("%d\n",list_tail->list_handle);
}
#endif
/*****************************************************************************/
void initial() {
  uint8_t i;
#ifdef DEBUG
  swap_out_num = 0;
  swap_in_num = 0;
#endif
  free_head = &free_list[0];
  for(i = 0; i < MAX_PM_PAGE; ++i) {
    free_list[i].next_p = &free_list[i+1];
  }
  free_list[MAX_PM_PAGE].next_p = NULL;
  list_head.list_handle = 0;    
  list_tail = &list_head;
  for (i = 0; i < MAX_VM_PAGE; ++i) {
    map[i] = NEVERUSE;
  } 
  for(i = 0; i < MAX_PM_PAGE; ++i) {
    freepages[i] = i;
  }
  vm_initial();
  cm_initial();

}             
/*****************************************************************************/
inline void * compute_w_addr(vir_addr) {
  uint16_t phy_addr;

#ifdef DEBUG
  ++access_num;
#endif

#ifndef TRANS
 uint8_t i;
 handle h = (uint8_t)(vir_addr>>AD_SHIFT);
    //map[h] = NEVERUSE means this page haven't been used yet
    //i.e. it's not in compressed area
  if(map[h] == NEVERUSE) {
      for(i = 0; i < MAX_PM_PAGE; ++i) {
	       if(freepages[i] != NEVERUSE) {
	          map[h] = freepages[i];
	          freepages[i] = NEVERUSE;
	          break;
	         }
      }
      if(i == MAX_PM_PAGE) {     //no free pages
	      map[h] = swap_out();
      }
      phy_addr =  vir_to_phy(vir_addr);
      add_to_list(h);
    } 
    else if (map[h] >= MAX_PM_PAGE) {
			swap_in(h);
			phy_addr = vir_to_phy(vir_addr);
			add_to_list(h);
    } else {
			phy_addr = vir_to_phy(vir_addr);
			replace_to_list(h);
    }

#else
   phy_addr = vir_to_phy(vir_addr);
#endif
    return pm_start + phy_addr;
}


/*****************************************************************************/
inline void * compute_r_addr(vir_addr) {
	uint16_t phy_addr;
#ifdef DEBUG
	++access_num;
#endif
#ifndef TRANS
	handle h = (uint8_t)((vir_addr) >> AD_SHIFT);
	if (map[h] >= MAX_PM_PAGE) {
		swap_in(h);
		phy_addr = vir_to_phy(vir_addr);
		add_to_list(h);
	} else {
		phy_addr =  vir_to_phy(vir_addr);
		replace_to_list(h);
	}
#else
	phy_addr = vir_to_phy(vir_addr);
#endif
	return pm_start + phy_addr;
}

/*****************************************************************************/
uint8_t read_handle8(uint16_t vir_addr) {
	return *(uint8_t *)compute_r_addr(vir_addr);
}

/*****************************************************************************/
uint16_t read_handle16(uint16_t vir_addr) {
	return *(uint16_t *)compute_r_addr(vir_addr);
}

/*****************************************************************************/
uint32_t read_handle32(uint16_t vir_addr) {
	return *(uint32_t *)compute_r_addr(vir_addr);
}

/*****************************************************************************/
int8_t read_handle8s(uint16_t vir_addr) {
	return *(int8_t *)compute_r_addr(vir_addr);
}

/*****************************************************************************/
int16_t read_handle16s(uint16_t vir_addr) {
	return *(int16_t *)compute_r_addr(vir_addr);
}

/*****************************************************************************/
int32_t read_handle32s(uint16_t vir_addr) {
	return *(int32_t *)compute_r_addr(vir_addr);
}
/*****************************************************************************/
void write_handle8(uint16_t vir_addr, uint8_t x) {
	*(uint8_t *)compute_w_addr(vir_addr) = x;
}

/*****************************************************************************/
void write_handle16(uint16_t vir_addr, uint16_t x) {
	*(uint16_t *)compute_w_addr(vir_addr) = x;
}

/*****************************************************************************/
void write_handle32(uint16_t vir_addr, uint32_t x) {
	*(uint32_t *)compute_w_addr(vir_addr) = x;
}

/*****************************************************************************/
void write_handle8s(uint16_t vir_addr, int8_t x) {
	*(int8_t *)compute_w_addr(vir_addr) = x;
}

/*****************************************************************************/
void write_handle16s(uint16_t vir_addr, int16_t x) {
	*(int16_t *)compute_w_addr(vir_addr) = x;
}

/*****************************************************************************/
void write_handle32s(uint16_t vir_addr, int32_t x) {
	*(int32_t *)compute_w_addr(vir_addr) = x;
}
/*****************************************************************************/
// check whether handle h is in uncompressed area
// if not, bring it into uncompressed area
void check_handle(handle h) {
    uint8_t i;

  	if(map[h] == NEVERUSE) {                //this page hasn't been used yet
    	for(i = 0; i < MAX_PM_PAGE; ++i) {
      	if(freepages[i] != NEVERUSE) {
					map[h] = freepages[i];
					freepages[i] = NEVERUSE;
				  break;
        }
      }
      if(i == MAX_PM_PAGE) {         //no free pages
        map[h] = swap_out();
      }
      add_to_list(h);
    } else {
	  	if(map[h] >= MAX_PM_PAGE) {     //handle h is in compressed area
    		swap_in(h);                  //bring h into uncompressed area
    		add_to_list(h);
  		}
  		else {
    		replace_to_list(h);
  		}
		}
}

/*****************************************************************************/
		
//end of  SOLU1
	
